#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022 "Devious 3" June 27, 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};


//config
#define ITERATIONS 200
#define EPSILON 1e-2
#define EPSILON_SCALE 1.03
//#define DISTANCE_STEP 1.
#define STEP_DITHER .26763
#define VOLUME_EPSILON 1e-1
#define LIGHTING_EPSILON 1e-4

//#define LIGHT_IMPORTANCE .01
#define LIGHT_DIRECTION normalize(vec3(-.483,1,0.283))
#define LIGHT_SIZE .01
#define LIGHT_COLOR vec3(.9,.83,.75)*5.
#define LIGHT_RESCALE_MAX 1e5
#define LIGHT_CAUSTICS 40.
#define LIGHT_MASK(p) 1.


//#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 1
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

float twave(float x) {
	return abs(fract(x)-.5)*2.;
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))





//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	/*animation ITER=100, EPS_SCALE=1.07
	return mat3(30,5.-55*pow(twave(time/60.),.5),30-time*4.,
			0,sin(time/60.*PI*2.)*PI*.3,PI,
			1.,0,0);*/
	/*screenshot 0, ITER=200, EPS_SCALE=1.03
	return mat3(243,-41,264.5,
			0,PI*.1,PI*.7,
			.7,0,0);*/
	/*screenshot 1*/
	return mat3(1.5,20,1.5,
			0,-PI*.5,0,
			6.,20,0);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//floors of spheres that turn into scattered cubes the farther down you go
	vec3 mp = mod(abs(p),vec3(3,10,3))-vec3(1.5,5,1.5),
		fp = floor(abs(p)/vec3(3,10,3));
	float ht = pow(min(1.,fp.y/5.),2.), iht = 1.-ht;
	mp.y += (twave(fp.x*.27893+fp.z*.368324)-.5)*ht*6.;
	mp = abs(mp);
	vec2 geo = vec2(max(p.y,length(max(mp-ht*1.5,0.))-iht),0);
	geo.x = min(geo.x,max(iht*5.-mp.y,2.-max(mp.x,mp.z)));

	//return vec2(1e6,0);
	return geo;
}
vec2 geometryDetail(vec3 p) {
	return geometry(p);
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;*/
mat2x3 material(vec3 p, float mat) {
	float rnd = twave(float(SFRAME)/PI+dot(abs(p),vec3(278.98724,56.762,761.2837)));
	if (rnd < 0.5) return mat2x3(vec3(.1), .8,0,1);//diffuse
	return mat2x3(vec3(.95), 0.,0,1);//reflective
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	float slen = length(d-normalize(vec3(.2293,1,-.7872)));
	return (mix(vec3(.9,.3,.1),vec3(.1,.2,.8),clamp(d.y*d.y+slen*.4,0.,1.))+
			vec3(1,.95,.6)/(.01+max(0.,slen-.1)*10.))*step(-3.,p.y);
	/*(vec3(1,.95,.6)*100.*max(1.-slen*2.,0.)+
		vec3(.56,.7,.9)*max(1.-slen*.4,0.))*step(-3.,p.y);*/
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#else
	time = 0.;
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;

	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;
	for (int i = 0; i < ITERATIONS; i++) {
		vec2 geo = geometry(rp);
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material
			if (random.w < mat[1].z) {//stochastic opacity
				
				#ifdef LIGHT_IMPORTANCE
				if (l.w < 0.) {	//failed importance cast
					c = vec3(0);
					break;
				}
				#endif
			
				rp -= rd*epsilon*2.;//back up ray
				c *= mat[0];//apply material color
				l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
				random = hash(uvec4(abs(rp.xyz)*10000.,SFRAME))*I2F;
				vec3 nrm = normal(rp);

				#ifdef LIGHT_IMPORTANCE
				if (random.w < LIGHT_IMPORTANCE) {//directional light importance sampling
					//brdf
					float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
						brdf = clamp((dot(LIGHT_DIRECTION,reflect(rd,nrm))-edge)/(1.-min(1.,edge)),0.,1.);
					rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
					
					if (dot(rd,nrm) < -.2 || brdf < 1e-5) {
						c = vec3(0);
						break;
					}
					c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
					if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
					l.w = -1.;//mark ray important
					
				} else {
				#endif
				
				//randomly bounce ray off microsurface bump
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));
				#ifdef LIGHT_IMPORTANCE
				}
				#endif
				
				geo.x = 0.;
			} else {
				geo.x = VOLUME_EPSILON;
			}
		} else {
			#ifdef DISTANCE_STEP
			geo.x *= DISTANCE_STEP;
			#endif
		}
		
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.w = fract(random.w+stepSz*2763.861234);
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*LIGHT_MASK(rp)*(1.+LIGHT_CAUSTICS)/LIGHT_IMPORTANCE;
	} else {
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif
	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
